@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 """
 discord.ext.commands
 ~~~~~~~~~~~~~~~~~~~~~
@@ -10,11 +8,13 @@ An extension module to facilitate creation of bot commands.
 :license: MIT, see LICENSE for more details.
 """
 
-from .bot import Bot, when_mentioned, when_mentioned_or
-from .context import Context
-from .core import *
-from .errors import *
-from .help import *
+from .bot import *
+from .cog import *
+from .context import *
 from .converter import *
 from .cooldowns import *
-from .cog import *
+from .core import *
+from .errors import *
+from .flags import *
+from .help import *
+from .parameters import *
